import logging
from typing import List, Tuple

from acedeploy.core.model_configuration import SolutionConfig
from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_prepostdeploymentsteps import (
    PreOrPostDeploymentScriptTarget,
    PreOrPostDeploymentScriptsExecutionOptions,
)
from acedeploy.extensions import persist_views_table
from acedeploy.services.db_compare_service import DbCompareClient
from acedeploy.services.deploy_service import DeployService
from acedeploy.services.git_service import GitClient
from acedeploy.services.metadata_service import MetadataService
from acedeploy.services.policy_service import PolicyService
from aceservices.snowflake_service import SnowClient, SnowClientConfig
from acedeploy.services.solution_service import SolutionClient
from acedeploy.services.dependency_parser import DependencyParser
from acedeploy.core.model_sql_entities import DbObjectType
from aceutils import database_management
from aceutils.action_generator import (
    generate_actions_from_git_diff,
    generate_actions_from_solution,
)
from aceutils.general_util import save_action_json, save_action_list
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def configure() -> SolutionConfig:
    """
    Load the current configuration from the path set in ACEDEPLOY_CONFIG_PATH.

    Returns:
        Configuration object
    """
    return SolutionConfig()


def execute_deployment(config: SolutionConfig) -> None:
    """
    Execute the deployment defined in the given `config`.

    Use `configure()` generate the config object.
    """
    target_clone_created = False

    try:
        # prepare clients
        snow_client_account = SnowClient(
            SnowClientConfig.get_from_solution_config(config)
        )
        database_management.create_database(
            config.database_meta,
            snow_client_account,
            replace_if_exists=True,
            drop_schema_public=True,
            db_retention_time_in_days=config.target_db_retention_time,
        )
        snow_client_meta = SnowClient(
            SnowClientConfig.get_from_solution_config(config, config.database_meta)
        )
        snow_client_target = SnowClient(
            SnowClientConfig.get_from_solution_config(config, config.database_target)
        )
        solution_client = SolutionClient(
            project_folder=config.project_folder,
            predeployment_folders=config.pre_deployment_folders,
            postdeployment_folders=config.post_deployment_folders,
            predeployment_settings=config.pre_deployment_settings,
            postdeployment_settings=config.post_deployment_settings,
            prepostdeployment_filter_list=config.prepostdeployment_filter_list,
            config_schema_list=config.schema_list,
            config_project_folder_filter=config.project_folder_filter,
            disabled_object_types=config.disabled_object_types,
        )

        log.info(f"SET deployment mode [ '{config.deployment_mode}' ]")

        # deploy to meta database
        if config.ignore_git_information is True:
            solution_client.load_solution(
                string_replace_dict=config.string_replace,
                pre_and_postdeployment_execution=PreOrPostDeploymentScriptsExecutionOptions.ALL,
            )
            dependency_client = DependencyParser(solution_client)
            dependency_client.build_full_dependency_graph()
            log.info(f"DEPLOY full solution TO meta db [ '{config.database_meta}' ]")
            action_list_meta = deploy_to_meta_database(
                solution_client,
                dependency_client,
                snow_client_meta,
                config,
                "solution",
                config.parallel_threads,
            )
        else:
            git_client = GitClient(
                config.solution_root_path, config.git_tag_regex, config.deployment_mode
            )
            git_changed_file_list = git_client.get_diff_filelist()
            solution_client.load_solution(
                git_changed_file_list=git_changed_file_list,
                string_replace_dict=config.string_replace,
                pre_and_postdeployment_execution=PreOrPostDeploymentScriptsExecutionOptions.GIT,
            )
            dependency_client = DependencyParser(solution_client)
            dependency_client.build_full_dependency_graph()
            log.info(f"DEPLOY git diff TO meta db [ '{config.database_meta}' ]")
            action_list_meta = deploy_to_meta_database(
                solution_client,
                dependency_client,
                snow_client_meta,
                config,
                "git",
                config.parallel_threads,
            )

        # create a clone
        if config.deploy_to_clone:
            log.info(
                f"CREATE TARGET database [ '{config.database_target}' ] as CLONE of [ '{config.clone_source_database}' ] with mode [ '{config.deploy_to_clone_mode}' ]"
            )
            pre_or_postdeployment_steps_will_be_executed = (
                len(
                    [
                        s
                        for s in solution_client.postdeployment_steps
                        + solution_client.predeployment_steps
                        if s.execute_step
                    ]
                )
                > 0
            )
            create_target_clone(
                snow_client_account,
                action_list_meta,
                config,
                pre_or_postdeployment_steps_will_be_executed,
            )
            target_clone_created = True

        # deploy to target database
        if config.deployment_mode.casefold() != "validate":
            log.info(
                f"START solution deployment TO target db [ '{config.database_target}' ]."
            )
            (
                action_list,
                predeployment_steps,
                postdeployment_steps,
            ) = deploy_to_target_database(
                snow_client_target,
                snow_client_meta,
                snow_client_account,
                solution_client,
                dependency_client,
                config,
                not config.abort_on_data_loss,
                config.deployment_mode.casefold() == "release",
            )

            if config.action_summary_path != "":
                log.info(f"SAVE ACTION SUMMARY TO [ '{config.action_summary_path}' ]")
                save_action_list(
                    config.action_summary_path,
                    action_list,
                    predeployment_steps,
                    postdeployment_steps,
                )

            if config.action_json_path != "":
                log.info(f"SAVE ACTION JSON TO [ '{config.action_json_path}' ]")
                save_action_json(
                    config.action_json_path,
                    action_list,
                    config.database_target,
                    config.snow_role,
                )

            # set flags in persist table
            if config.persist_views_table:
                persist_views_table.set_persist_tags(
                    config.persist_views_table,
                    action_list,
                    config.database_target,
                    snow_client_account,
                )

    except Exception as err:
        log.error(str(err))
        raise err

    finally:
        database_management.drop_database_and_set_retention_time(
            database_name=config.database_meta,
            snow_client=snow_client_account,
            retention_time_in_days=config.temp_db_retention_time,
        )
        if (
            target_clone_created
            and config.deploy_to_clone
            and config.deploy_to_clone_mode
            and config.drop_clone_after_run
            and (config.clone_source_database.lower() != config.database_target.lower())
        ):
            database_management.drop_database_and_set_retention_time(
                database_name=config.database_target,
                snow_client=snow_client_account,
                retention_time_in_days=config.temp_db_retention_time,
            )


def deploy_to_meta_database(
    solution_client: SolutionClient,
    dependency_client: DependencyParser,
    snow_client: SnowClient,
    config: SolutionConfig,
    mode: str,
    parallel_threads: int,
) -> List[List[List[DbObjectAction]]]:
    """
    Deploy entire solution or git diff to meta database.

    Used to validate the sql code and to initialize the meta db for comparison with the target db in a later step.

    Args:
        `solution_client`: Collection of solution items
        `dependency_client`: (Unordered) graph of solution items
        `snow_client`: Connection to Snowflake (must be scoped to meta database)
        `mode`: 'git' or 'solution'. Flag wether we are deploying changes based off git or all solution items
        `parallel_threads`: Number of parallel threads during deployment

    Returns:
        Nested list of actions that were performed during the deployment.
        The list is nested and ordered to allow parallel deployment of objects which do not share dependencies.
        Refer to `deploy_client.start_deployment()` for details.
    """
    action_list = []
    deploy_client = DeployService(solution_client, dependency_client, snow_client, config)
    log.info(f"START validation TO meta db [ '{snow_client.database}' ]")
    if mode == "git":
        action_list = generate_actions_from_git_diff(solution_client, always_create_tags=config.always_create_tags)
    elif mode == "solution":
        action_list = generate_actions_from_solution(solution_client)
    else:
        raise ValueError(f"ERROR invalid meta deployment mode [ '{mode}' ]")

    deploy_client.execute_predeployment(target=PreOrPostDeploymentScriptTarget.META)
    action_list = deploy_client.start_deployment(
        action_list, True, parallel_threads=parallel_threads
    )
    deploy_client.execute_postdeployment(target=PreOrPostDeploymentScriptTarget.META)
    log.info(f"SUCCESS validation TO meta db [ '{snow_client.database}' ]")

    return action_list


def deploy_to_target_database(
    snow_client_target: SnowClient,
    snow_client_meta: SnowClient,
    snow_client_account: SnowClient,
    solution_client: SolutionClient,
    dependency_client: DependencyParser,
    config: SolutionConfig,
    allow_dataloss: bool,
    rollback_enabled: bool,
) -> Tuple[List[DbObjectAction], List[str], List[str]]:
    """
    Deploy changes to the target database.

    Steps:
        1) Execute predeployment scripts on target database
        2) Compare meta database and target database and determine required actions
        3) Execute required actions on target database
        4) Execute postdeployment scripts on target database

    Args:
        `snow_client_target`: Connection to Snowflake (must be scoped to target database)
        `snow_client_meta`: Connection to Snowflake (must be scoped to meta database)
        `snow_client_account`: Connection to Snowflake (must not be scoped to a database)
        `solution_client`: Collection of solution items
        `dependency_client`: (unordered) graph of solution items
        `config`: Deployment configuration
        `allow_dataloss`: Flag wether deployment will be performed if dataloss in a table is detected
        `rollback_enabled`: Flag wether roll back will be performed on failed deployment

    Returns:
        Nested list of actions that were performed during the deployment.
        The list is nested and ordered to allow parallel deployment of objects which do not share dependencies.
        Refer to `deploy_client.start_deployment()` for details.
    """
    deploy_client = DeployService(
        solution_client,
        dependency_client,
        snow_client_target,
        config,
        snow_client_meta=snow_client_meta,
        snow_client_account=snow_client_account,
        release_name=config.release_name,
        temp_db_retention_time=config.temp_db_retention_time,
    )
    predeployment_steps = deploy_client.execute_predeployment(
        target=PreOrPostDeploymentScriptTarget.TARGET
    )
    action_list, policy_assignments_handler = generate_actions_from_db_diff(
        snow_client_meta, snow_client_target, solution_client, config
    )

    policy_service = None
    if config.handle_policy_assignments:
        log.info(f'START handling policy assignments.')
        policy_service = PolicyService(
            policy_assignments_handler, 
            config.policy_assignments_deployment_database, 
            config.policy_assignments_config_file_path, 
            config.policy_assignments_project, 
            config.policy_assignments_info_output_folder_path,
            config.policy_assignments_repo_path
        )
        policy_service.get_policy_assignments_info()

    deploy_client.start_deployment(
        action_list=action_list,
        is_meta_deployment=False,
        allow_dataloss=allow_dataloss,
        rollback_enabled=rollback_enabled,
        schema_list=config.schema_list,
        parallel_threads=config.parallel_threads,
        autoresume_tasks=config.autoresume_tasks,
        policy_service=policy_service,
        execute_in_parallel_sessions=config.execute_in_parallel_sessions,
    )

    if config.handle_policy_assignments:
        if config.policy_assignments_save_info:
            policy_service.save_policy_assignments_info()
        if config.policy_assignments_create_azure_pull_request_comments:
            policy_service.create_azure_devops_pr_comment_from_policy_assignments_info()


    postdeployment_steps = deploy_client.execute_postdeployment(
        target=PreOrPostDeploymentScriptTarget.TARGET
    )
    return action_list, predeployment_steps, postdeployment_steps


def generate_actions_from_db_diff(
    snow_client_meta: SnowClient,
    snow_client_target: SnowClient,
    solution_client: SolutionClient,
    config: SolutionConfig,
) -> List[DbObjectAction]:
    """
    Compare meta and target database and generate list of actions.

    The meta database corresponds to the desired state of the target.
    By comparing both databases a list of required actions is generated.

    Args:
        `snow_client_meta`: Connection to Snowflake (must be scoped to meta database)
        `snow_client_target`: Connection to Snowflake (must be scoped to target database)
        `solution_client`: Collection of solution items
        `config`: Deployment configuration

    Returns:
        Unordered list of actions
    """
    metadata_service_metadb = MetadataService(
        snow_client=snow_client_meta,
        snow_edition=config.snow_edition,
        disabled_object_types=config.disabled_object_types,
        object_options=config.object_options,
        workarounds_options=config.workarounds
    )
    metadata_service_metadb.get_all_metadata(
        schema_list=config.schema_list, 
        parallel_threads=config.parallel_threads,
        check_for_language_restrictions=True,
    )
    schemas_on_meta = [s.name for s in metadata_service_metadb.schemas]
    schema_filter_target = {"whitelist": schemas_on_meta}

    metadata_service_targetdb = MetadataService(
        snow_client=snow_client_target,
        snow_edition=config.snow_edition,
        disabled_object_types=config.disabled_object_types,
        object_options=config.object_options,
        workarounds_options=config.workarounds
    )
    metadata_service_targetdb.get_all_metadata(
        schema_filter_target,
        config.parallel_threads,
        get_policies_on_objects_legacy=config.reapply_existing_policies,
    )

    log.info("GET information schema DIFF")

    compare_client = DbCompareClient(
        metadata_service_metadb, metadata_service_targetdb, solution_client
    )
    compare_client.get_add_actions()
    compare_client.get_alter_actions()
    if config.object_options.get(DbObjectType.PROCEDURE, {}).dropOverloadedObjects or config.object_options.get(DbObjectType.FUNCTION, {}).dropOverloadedObjects:
        compare_client.get_drop_overloaded_actions(config.object_options.get(DbObjectType.PROCEDURE, {}).dropOverloadedObjects, config.object_options.get(DbObjectType.FUNCTION, {}).dropOverloadedObjects)

    if config.drop_target_objects:
        if config.ignore_git_information:
            compare_client.get_drop_actions_from_solution()
        else:
            compare_client.get_drop_actions_from_git()

    return compare_client.action_list, compare_client.policy_assignments_handler


def create_target_clone(
    snow_client_account: SnowClient,
    action_list_meta: List[List[List[DbObjectAction]]],
    config: SolutionConfig,
    pre_or_postdeployment_steps_will_be_executed: bool,
) -> None:
    """
    Create a clone of a database to be used as the target database.

    There are different options on how exactly the clone is created, as set in
    `config.deploy_to_clone_mode`: `minimal`, `schemalist`, `full`.

    In mode `minimal`, if `pre_or_postdeployment_steps_will_be_executed` is true, mode
    `schemalist` is used instead. If cloning fails in mode `minimal`, cloning
    automatically falls back to mode `schemalist`.

    Args:
        `snow_client_meta`: Connection to Snowflake (must be scoped to meta db)
        `action_list_meta`: Ordered list of actions, as returned by `deploy_client.start_deployment()` after the meta deployment
        `config`: Deployment configuration
        `pre_or_postdeployment_steps_will_be_executed`: Flag wether the deployment will contain pre- or postdeployment steps.
    """
    mode = config.deploy_to_clone_mode.lower()
    if mode == "minimal" and pre_or_postdeployment_steps_will_be_executed:
        log.warning(
            "Minimal cloning not possible if pre- or postdeployment steps will be executed. Will create full database clone instead. This will take longer."
        )
        mode = "schemalist"
    if mode == "minimal":
        try:
            database_management.create_clone_by_ordered_action_list(
                config.clone_source_database,
                config.database_target,
                action_list_meta,
                snow_client_account,
                replace_if_exists=False,
                db_retention_time_in_days=config.target_db_retention_time,
                policy_assignments_role=config.policy_assignments_role
            )
        except Exception as err:
            log.warning(
                "Minimal cloning was not successful. Will create full database clone instead. This will take longer."
            )
            log.info(f"Reason for failed minimal clone step: {str(err)}")
            database_management.clone_database_by_schemas(
                config.clone_source_database,
                config.database_target,
                config.schema_list,
                snow_client_account,
                replace_if_exists=True,
                parallel_threads=config.parallel_threads,
                db_retention_time_in_days=config.target_db_retention_time,
            )
    elif mode == "schemalist":
        database_management.clone_database_by_schemas(
            config.clone_source_database,
            config.database_target,
            config.schema_list,
            snow_client_account,
            replace_if_exists=False,
            parallel_threads=config.parallel_threads,
            db_retention_time_in_days=config.target_db_retention_time,
            ignore_hybrid_tables=config.clone_ignore_hybrid_tables,
            create_hybrid_tables_from_ddl=config.clone_create_hybrid_tables_from_ddl,
        )
    elif mode == "full":
        database_management.clone_database(
            config.clone_source_database,
            config.database_target,
            snow_client_account,
            replace_if_exists=False,
        )
    else:
        raise ValueError(
            f"Invalid value for config.deploy_to_clone_mode. Must be one of (minimal, schemalist, full). Got '{config.deploy_to_clone_mode}'."
        )



import re
from typing import Union


def map_datatype_name_to_default(datatype: str) -> str:
    """
    Return the default representation of a snowflake datatype.

    See also https://docs.snowflake.com/en/sql-reference/intro-summary-data-types.html

    Examples:
        TEXT -> VARCHAR
        INT -> NUMBER
        NUMBER(38,5) -> NUMBER
        VARCHAR(10) -> VARCHAR
    """
    # remove precision like (38,0) and strip whitespace
    datatype_clean = re.sub(r"\s*\([\d\s,]+\)", "", datatype.upper().strip())

    if datatype_clean in (
        "NUMBER",
        "DECIMAL",
        "NUMERIC",
        "INT",
        "INTEGER",
        "BIGINT",
        "SMALLINT",
    ):
        return "NUMBER"

    if datatype_clean in ("VARCHAR", "CHAR", "CHARACTER", "STRING", "TEXT"):
        return "VARCHAR"

    if datatype_clean in ("BINARY", "VARBINARY"):
        return "BINARY"

    if datatype_clean in ("DATETIME", "TIMESTAMP", "TIMESTAMP_NTZ"):
        return "TIMESTAMP_NTZ"

    if datatype_clean in (
        "FLOAT",
        "FLOAT4",
        "FLOAT8",
        "DOUBLE",
        "DOUBLE PRECISION",
        "REAL",
    ):
        return "FLOAT"

    return datatype


def string_or_bool_to_bool(v: Union[str, bool]) -> bool:
    """
    Given a boolean, return that boolean.
    Given a string,
    """
    vstr = str(v).lower()
    if vstr == "true":
        return True
    elif vstr == "false":
        return False
    else:
        raise ValueError(f"Given value '{v}' can not be interpreted as boolean.")


def compare_datatype_ignore_varchar_maximum_length(
    datatype_1: str, datatype_2: str
) -> bool:
    """
    This function has the purpose of ignoring the maximum length of the VARCHAR datatype during datatype comparison.
    Introduced by the Change Bundle 2024_08, Snowflake omits the max length of the data type VARCHAR in specific output cases.
    """

    if (datatype_1 == "VARCHAR(16777216)" and datatype_2 == "VARCHAR") or (
        datatype_1 == "VARCHAR" and datatype_2 == "VARCHAR(16777216)"
    ):
        return True
    else:
        return datatype_1 == datatype_2



import abc
import os
from enum import Enum


class SecretService(abc.ABC):
    """
    Base class for all secret retrieval services
    """

    @abc.abstractmethod
    def get_secret(self, secret_name: str):
        pass


class SecretServiceType(Enum):
    """
    The type of secret service to use
    """

    ENVIRONMENT = "ENVIRONMENT"


class EnvironmentSecretService(SecretService):
    """
    Retrieves secrets from environment variables
    """

    key_map: dict = {}

    def __init__(self, key_map: dict = None):
        """
            Inits a new environment secret service
        Args:
            key_map: dict - optional dictionary that maps secret keys to environment variables
        """
        self.key_map = {} if key_map is None else key_map

    def get_secret(self, secret_name: str):
        """
            Retrieves the environment variable by name
            First tries to lookup key in key mapping
            secret_name is case-sensitive
        Args:
            secret_name: str - the name of the secret to retrieve
        """
        key = self.key_map[secret_name] if secret_name in self.key_map else secret_name
        secret_value = os.getenv(key)

        if secret_value is None:
            raise ValueError(f"Environment variable {secret_name} not found")

        return secret_value



import json
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, Union, DefaultDict
from collections import defaultdict  

import aceutils.file_util as file_util
import aceutils.string_util as string_util
import aceutils.parallelization_util as parallelization_util
from acedeploy.core.model_instance_objects import InstanceObject
from acedeploy.core.model_sql_entities import DbObjectType
from acedeploy.core.model_configuration import ObjectOption
from aceservices.snowflake_service import SnowClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class MetadataService:
    """
    Class to query metadata information from snowflake.
    An instance of this class is scoped to a single database (defined by the given snow_client)
    """

    def __init__(
        self,
        snow_client: SnowClient,
        snow_edition: str = "Enterprise",
        disabled_object_types: List[DbObjectType] = [], # TODO: consider removing this and referencing the information from object_options instead
        object_options: Dict[DbObjectType, ObjectOption] = None,
        workarounds_options: Dict = {}
    ):
        """
            Inits a metadata service.
        Args:
            snow_client: SnowClient - provides connection a snowflake database
            snow_edition: str - Edition of the Snowflake account, "Enterprise" or "Standard"
            disabled_object_types: List[DbObjectType] - List of sql object types for which no metadata should be queried
            object_options: Dict[DbObjectType, ObjectOption] - Dict of object options (e.g. whether tags should be queried for each type)
        """
        self._snow_client = snow_client
        self._snow_edition = snow_edition
        self._disabled_object_types = disabled_object_types
        self._object_options = object_options
        self._workarounds_options = workarounds_options
        module_root_folder = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..")
        )
        self._sql_files_path = os.path.abspath(
            os.path.join(module_root_folder, "resources", "sql")
        )
        self.all_objects = []        
        self.all_objects_dict_by_id = None
        self.database_name = None
        self.check_for_language_restrictions = None

    def get_all_objects_subset(self, object_type: DbObjectType):
        return [o for o in self.all_objects if o.object_type == object_type]

    @property
    def schemas(self):
        return self.get_all_objects_subset(DbObjectType.SCHEMA)

    @property
    def tables(self):
        return self.get_all_objects_subset(DbObjectType.TABLE)

    @property
    def externaltables(self):
        return self.get_all_objects_subset(DbObjectType.EXTERNALTABLE)

    @property
    def views(self):
        return self.get_all_objects_subset(DbObjectType.VIEW)

    @property
    def materializedviews(self):
        return self.get_all_objects_subset(DbObjectType.MATERIALIZEDVIEW)

    @property
    def functions(self):
        return self.get_all_objects_subset(DbObjectType.FUNCTION)

    @property
    def procedures(self):
        return self.get_all_objects_subset(DbObjectType.PROCEDURE)

    @property
    def fileformats(self):
        return self.get_all_objects_subset(DbObjectType.FILEFORMAT)

    @property
    def stages(self):
        return self.get_all_objects_subset(DbObjectType.STAGE)

    @property
    def streams(self):
        return self.get_all_objects_subset(DbObjectType.STREAM)

    @property
    def tasks(self):
        return self.get_all_objects_subset(DbObjectType.TASK)

    @property
    def pipes(self):
        return self.get_all_objects_subset(DbObjectType.PIPE)

    @property
    def sequences(self):
        return self.get_all_objects_subset(DbObjectType.SEQUENCE)

    @property
    def maskingpolicies(self):
        return self.get_all_objects_subset(DbObjectType.MASKINGPOLICY)

    @property
    def rowaccesspolicies(self):
        return self.get_all_objects_subset(DbObjectType.ROWACCESSPOLICY)
    
    @property
    def dynamictables(self):
        return self.get_all_objects_subset(DbObjectType.DYNAMICTABLE)
    
    @property
    def hybridtables(self):
        return self.get_all_objects_subset(DbObjectType.HYBRIDTABLE)
    
    @property
    def networkrules(self):
        return self.get_all_objects_subset(DbObjectType.NETWORKRULE)

    @property
    def tags(self):
        return self.get_all_objects_subset(DbObjectType.TAG)


    def get_object_by_object(self, obj: InstanceObject) -> InstanceObject:
        """
        Return the matching InstanceObject for the given obj.
        Compares the id of the object.
        """
        return self.get_object_by_id(obj.id)

    def get_object_by_id(self, object_id: str) -> InstanceObject:
        """
        Return the InstanceObject for the given object_id.
        """
        result = self.all_objects_dict_by_id[object_id.replace('"', "")]
        if len(result) == 0:
            log.debug(
                f"OBJECT with [ id='{object_id}' ] NOT in FOUND in MetadataService"
            )
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with [ id='{object_id}' ] NOT UNIQUE in MetadataService. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]
        
    def _populate_all_objects_dict_by_id(self) -> None:
        self.all_objects_dict_by_id = defaultdict(list)
        for obj in self.all_objects:
            self.all_objects_dict_by_id[obj.id].append(obj)

    def get_all_metadata(
        self,
        schema_list: Dict[str, List[str]],
        parallel_threads: int = 1,
        get_policies_on_objects_legacy: bool = False,
        check_for_language_restrictions: bool = None, 
    ) -> None:
        """
        Get all metadata information (schemas, tables+views and other objects) from the database.
        """
        log.info(f"QUERY object information ON [ '{self._snow_client.database}' ].")

        mapping = {
            DbObjectType.SCHEMA: self._get_metadata_schemas,
            DbObjectType.FUNCTION: self._get_metadata_functions,
            DbObjectType.PROCEDURE: self._get_metadata_procedures,
            DbObjectType.FILEFORMAT: self._get_metadata_fileformats,
            DbObjectType.STREAM: self._get_metadata_streams,
            DbObjectType.TASK: self._get_metadata_tasks,
            DbObjectType.SEQUENCE: self._get_metadata_sequences,
            DbObjectType.MATERIALIZEDVIEW: self._get_metadata_materializedviews,
            DbObjectType.TABLE: self._get_metadata_tables,
            DbObjectType.EXTERNALTABLE: self._get_metadata_externaltables,
            DbObjectType.VIEW: self._get_metadata_views,
            DbObjectType.STAGE: self._get_metadata_stages,
            DbObjectType.PIPE: self._get_metadata_pipes,
            DbObjectType.MASKINGPOLICY: self._get_metadata_maskingpolicies,
            DbObjectType.ROWACCESSPOLICY: self._get_metadata_rowaccesspolicies,
            DbObjectType.DYNAMICTABLE: self._get_metadata_dynamictables,
            DbObjectType.HYBRIDTABLE: self._get_metadata_hybridtables,
            DbObjectType.NETWORKRULE: self._get_metadata_networkrules,
            DbObjectType.TAG: self._get_metadata_tags,
        }

        self.database_name = self._snow_client.database
        self.check_for_language_restrictions = check_for_language_restrictions
        self._get_policies_on_objects_legacy = get_policies_on_objects_legacy

        schemas_to_query = self._get_filtered_schema_name_list(schema_list)

        if self._object_options is not None and any([getattr(oo, "manageRowAccessPolicyReferences", False) for oo in self._object_options.values()]):
            row_access_policy_references = self._get_row_access_policy_references(schemas_to_query, parallel_threads)

        if self._object_options is not None and any([getattr(oo, "manageMaskingPolicyReferences", False) for oo in self._object_options.values()]):
            masking_policy_references = self._get_masking_policy_references(schemas_to_query, parallel_threads)

        for object_type, metadata_function in mapping.items():
            if object_type not in self._disabled_object_types:
                log.info(f"QUERY object information FOR type [ '{object_type}' ]")
                metadata_list = metadata_function(schemas_to_query, parallel_threads)
                if self._object_options is not None and getattr(self._object_options[object_type], "manageRowAccessPolicyReferences", False):
                    metadata_list = self._extend_metadata_policyreferences(metadata_list, row_access_policy_references[object_type])
                if self._object_options is not None and getattr(self._object_options[object_type], "manageMaskingPolicyReferences", False):
                    metadata_list = self._extend_metadata_policyreferences(metadata_list, masking_policy_references[object_type])
                if self._object_options is not None:
                    metadata_list = self._update_metadata_using_object_options(metadata_list, getattr(self._object_options[object_type], "metadataOptions", {}))
                self.all_objects.extend(
                    [InstanceObject.factory(
                        object_type=object_type,
                        metadata_query_result=m,
                        object_options=self._object_options[object_type] if self._object_options is not None else None,
                    ) for m in metadata_list]
                )            
        self._populate_all_objects_dict_by_id()


    def _extend_metadata_policyreferences(self, metadata_list: List[Dict], policy_references: List[Dict]):
        for obj in metadata_list:
            refs = [
                ref for ref in policy_references if (
                    ref["REF_DATABASE_NAME"] == obj["DATABASE_NAME"]
                    and ref["REF_SCHEMA_NAME"] == obj["SCHEMA_NAME"]
                    and ref["REF_ENTITY_NAME"] == obj["OBJECT_NAME"]
                )
            ]
            if "policy_references" in obj:
                obj["policy_references"].extend(refs)
            else:
                obj["policy_references"] = refs
        return metadata_list

    @staticmethod
    def _update_metadata_using_object_options(metadata_list: List[Dict], metadata_options: Dict[str, Union[Dict[str, str], Dict[str, Dict[str, str]]]]):
        """
            Given a list of metadata, update certain fields according to the metadata options.
        """
        def _update_single_property(this_options: Dict, this_obj: any, this_property: str):
            for option_key, option_value in this_options.items():
                if option_key == "ignore":
                    if option_value:
                        MetadataService._set_ignore_value(this_obj, this_property)
                else:
                    raise ValueError(f"Unknown metadataOption '{option_key}'")

        for obj in metadata_list:
            for property, options in metadata_options.items():
                if property == "COLUMN_DETAILS":
                    for subproperty, suboptions in metadata_options[property].items():
                        for col in obj[property]:
                            _update_single_property(suboptions, col, subproperty)
                else:
                    _update_single_property(options, obj, property)

        return metadata_list


    @staticmethod
    def _set_ignore_value(metadata_dict: Dict, property: str) -> None:
        """
        Given a metadata dict and a property name, set the property to an "ignore" value (e.g. None or empty list/dict).
        Updates the dict in place.
        """
        if isinstance(metadata_dict.get(property), dict):
            metadata_dict[property] = {}
        elif isinstance(metadata_dict.get(property), list):
            metadata_dict[property] = []
        elif isinstance(metadata_dict.get(property), set):
            metadata_dict[property] = set()
        else:
            metadata_dict[property] = None


    def _get_row_access_policy_references(self, schemas_to_query, parallel_threads) -> Dict[DbObjectType, List[Dict]]:
        log.debug(f"QUERY row access policy references")
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(
            self._sql_files_path, "get_rowaccesspolicy_names.sql"
        )
        object_type = DbObjectType.ROWACCESSPOLICY
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_policy_references(metadata_list)

        # at this point, we only care about entries in the policy_references field
        # the actual policy metadata is fetched in another function
        row_access_policy_references_list = []
        for metadata_entry in metadata_list:
            row_access_policy_references_list.extend(metadata_entry["references"])

        row_access_policy_references = {}
        for object_type, object_option in self._object_options.items():
            if getattr(object_option, "manageRowAccessPolicyReferences", False):
                row_access_policy_references[object_type] = [ref for ref in row_access_policy_references_list if object_type == DbObjectType.get_object_type_for_policy_references(ref["REF_ENTITY_DOMAIN"])]
            else:
                row_access_policy_references[object_type] = []

        return row_access_policy_references

    def _get_masking_policy_references(self, schemas_to_query, parallel_threads) -> Dict[DbObjectType, List[Dict]]:
        log.debug(f"QUERY masking policy references")
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(
            self._sql_files_path, "get_maskingpolicy_names.sql"
        )
        object_type = DbObjectType.MASKINGPOLICY
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_policy_references(metadata_list)

        # at this point, we only care about entries in the policy_references field
        # the actual policy metadata is fetched in another function
        masking_policy_references_list = []
        for metadata_entry in metadata_list:
            masking_policy_references_list.extend(metadata_entry["references"])

        masking_policy_references = {}
        for object_type, object_option in self._object_options.items():
            if getattr(object_option, "manageMaskingPolicyReferences", False):
                masking_policy_references[object_type] = [ref for ref in masking_policy_references_list if object_type == DbObjectType.get_object_type_for_policy_references(ref["REF_ENTITY_DOMAIN"])]
            else:
                masking_policy_references[object_type] = []

        return masking_policy_references

    def _get_metadata_schemas(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_schemas.sql")
        object_type = DbObjectType.SCHEMA
        return self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

    def _get_metadata_functions(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_functions.sql")
        object_type = DbObjectType.FUNCTION
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )  # note: metadata details are currently not evaluated, see db_compare_service.py

        # check for disabled languages
        if self._object_options and self.check_for_language_restrictions:
            for udf in metadata_list:
                if udf["LANGUAGE"].upper() in self._object_options[object_type].disabledLanguages:
                    raise ValueError(f'The usage of language {udf["LANGUAGE"]} for User-Defined Functions is disabled! Detected in UDF: {udf["DATABASE_NAME"]}.{udf["SCHEMA_NAME"]}.{udf["OBJECT_NAME"]}')
        return metadata_list

    def _get_metadata_procedures(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_procedures.sql"
        )
        object_type = DbObjectType.PROCEDURE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )  # note: metadata details are currently not evaluated, see db_compare_service.py

        # check for disabled languages
        if self._object_options and self.check_for_language_restrictions:
            for procedure in metadata_list:
                if procedure["LANGUAGE"].upper() in self._object_options[object_type].disabledLanguages:
                    raise ValueError(f'The usage of language {procedure["LANGUAGE"]} for Stored Procedures is disabled! Detected in Stored Procedure: {procedure["DATABASE_NAME"]}.{procedure["SCHEMA_NAME"]}.{procedure["OBJECT_NAME"]}')
        return metadata_list

    def _get_metadata_fileformats(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_fileformats.sql"
        )
        object_type = DbObjectType.FILEFORMAT
        return self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

    def _get_metadata_streams(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_streams.sql")
        object_type = DbObjectType.STREAM
        return self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

    def _get_metadata_tasks(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_tasks.sql")
        object_type = DbObjectType.TASK
        return self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )  # note: metadata details are currently not evaluated, see db_compare_service.py

    def _get_metadata_sequences(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_sequences.sql")
        object_type = DbObjectType.SEQUENCE
        return self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

    def _get_metadata_materializedviews(self, schemas_to_query, parallel_threads):
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_materializedviews.sql"
        )
        object_type = DbObjectType.MATERIALIZEDVIEW
        metadata_list= self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )  # note: metadata details are currently not evaluated, see db_compare_service.py

        self._convert_column_details_to_dict(metadata_list)

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)
        return metadata_list

    def _get_metadata_tables(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_tables.sql")
        object_type = DbObjectType.TABLE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._get_constraint_metadata(
            schemas_to_query, parallel_threads, metadata_list
        )


        self._convert_column_details_to_dict(metadata_list)

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)


        return metadata_list

    def _get_metadata_externaltables(self, schemas_to_query, parallel_threads):
        
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_externaltables.sql"
        )
        object_type = DbObjectType.EXTERNALTABLE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        # TODO: evaluate if we want to include constraint metadata here (for now this is not required, because we the metadata of external tables is not evaluated, see db_compare_service.py)
        

        self._convert_column_details_to_dict(metadata_list)

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)
        
        
        return metadata_list

    def _get_metadata_views(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_views.sql")
        object_type = DbObjectType.VIEW
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        if self._get_policies_on_objects_legacy:
            metadata_list = self._query_policies_on_objects_legacy(metadata_list)

        self._convert_column_details_to_dict(metadata_list)
        
        if self._workarounds_options.get("snowflakeBugs", {}).get("2025-02-clonedViewsQuotes", {}).get("enabled", False):
            metadata_list = self._fix_view_definitions(metadata_list) # This fixes a Snowflake bug when retrieving the VIEW_DEFINITION from the INFORMATION_SCHEMA.VIEWS after cloning the schema that contains the view.

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)

        return metadata_list
    
    def _fix_view_definitions(self, metadata_list: List[dict]) -> List[dict]:
        """
        This fixes a Snowflake bug when retrieving the VIEW_DEFINITION from the INFORMATION_SCHEMA.VIEWS after cloning the schema that contains the view.
        """

        for n, metadata_view in enumerate(metadata_list):
                
            if 'VIEW_DEFINITION' in metadata_view:
                metadata_view['VIEW_DEFINITION'] = string_util.fix_view_definition(metadata_view['VIEW_DEFINITION'])

            metadata_list[n] = metadata_view

        return metadata_list

    def _get_metadata_stages(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_stages.sql")
        object_type = DbObjectType.STAGE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_stages(metadata_list)
        return metadata_list

    def _get_metadata_pipes(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_pipes.sql")
        object_type = DbObjectType.PIPE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_pipes(metadata_list)
        return metadata_list

    def _get_metadata_maskingpolicies(self, schemas_to_query, parallel_threads):
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(
            self._sql_files_path, "get_maskingpolicy_names.sql"
        )
        object_type = DbObjectType.MASKINGPOLICY
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_maskingpolicies(metadata_list)
        return metadata_list

    def _get_metadata_rowaccesspolicies(self, schemas_to_query, parallel_threads):
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(
            self._sql_files_path, "get_rowaccesspolicy_names.sql"
        )
        object_type = DbObjectType.ROWACCESSPOLICY
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._extend_metadata_rowaccesspolicies(metadata_list)
        return metadata_list
    
    def _get_metadata_dynamictables(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_dynamictables.sql")
        
        object_type = DbObjectType.DYNAMICTABLE

        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

        self._convert_column_details_to_dict(metadata_list)

        metadata_list = self._extend_metadata_dynamictables(metadata_list, object_type, schemas_to_query, parallel_threads)

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)

        return metadata_list
    
    def _extend_metadata_dynamictables(self, metadata_list: List[Dict], object_type: DbObjectType, schemas_to_query: list, parallel_threads: int) -> List[Dict]:
        """
        Dynamic tables metadata needs to be extended with metadata from "show dynamic tables".
        """
        template_path_extended = os.path.join(self._sql_files_path, "get_metadata_dynamictables_extended.sql")

        metadata_list_extended = self._query_metadata(
            object_type, template_path_extended, schemas_to_query, parallel_threads
        )

        used_keys = ['target_lag', 'warehouse', 'refresh_mode']

        metadata_list_ = {f"{metadata['DATABASE_NAME']}.{metadata['SCHEMA_NAME']}.{metadata['OBJECT_NAME']}": metadata for metadata in metadata_list}

        for extended_metadata in metadata_list_extended:

            extended_metadata_ = {used_key.upper(): extended_metadata[used_key] for used_key in used_keys}

            metadata_list_[f"{extended_metadata['database_name']}.{extended_metadata['schema_name']}.{extended_metadata['name']}"].update(extended_metadata_)

        metadata_list = list(metadata_list_.values())

        return metadata_list

    def _get_metadata_hybridtables(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_hybridtables.sql")
        object_type = DbObjectType.HYBRIDTABLE
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        metadata_list = self._get_constraint_metadata(
            schemas_to_query, parallel_threads, metadata_list
        )

        self._convert_column_details_to_dict(metadata_list)

        if self._object_options is not None and self._object_options[object_type].manageTagAssignments:
            metadata_list = self._extend_metadata_with_tags(metadata_list, parallel_threads, object_type)

        return metadata_list
    
    def _get_metadata_networkrules(self, schemas_to_query, parallel_threads):
        template_path = os.path.join(self._sql_files_path, "get_metadata_networkrules.sql")
        
        object_type = DbObjectType.NETWORKRULE

        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )

        self._convert_column_details_to_dict(metadata_list)

        metadata_list = self._extend_metadata_networkrules(metadata_list, object_type, schemas_to_query, parallel_threads)

        return metadata_list
    
    def _extend_metadata_networkrules(self, metadata_list: List[Dict], object_type: DbObjectType, schemas_to_query: list, parallel_threads: int) -> List[Dict]:
        """
        Network rules metadata needs to be extended with metadata from "show network rules".
        """

        for metadata_entry in metadata_list:
            query_template = f'DESCRIBE NETWORK RULE {metadata_entry["schema_name"]}.{metadata_entry["name"]}'
            metadata_result = self._snow_client.execute_query(query_template, use_dict_cursor=True)
            metadata_entry.update(metadata_result[0])
            # Update keys to uppercase in place
            for key in list(metadata_entry.keys()):  # Use list() to avoid runtime changes during iteration
                metadata_entry[key.upper()] = metadata_entry.pop(key)
            metadata_entry["OBJECT_NAME"] = metadata_entry["NAME"]

        return metadata_list
    
    def _get_metadata_tags(self, schemas_to_query, parallel_threads):
        if self._snow_edition == "Standard":
            return []
        template_path = os.path.join(self._sql_files_path, "get_metadata_tags.sql")
        object_type = DbObjectType.TAG
        metadata_list = self._query_metadata(
            object_type, template_path, schemas_to_query, parallel_threads
        )
        return metadata_list


    def _query_metadata(
        self,
        object_type: DbObjectType,
        template_path: str,
        schemas_to_query: List[str],
        parallel_threads: int = 1,
        database_name="",
    ) -> List[Dict]:
        """
        Query metadata for a given object type. Return raw result
        """
        log.debug(f"QUERY object information for TYPE [ '{str(object_type)}' ].")
        query_template = file_util.load(template_path)
        queries = [
            query_template.format(schema_name=schema, database_name=database_name)
            for schema in schemas_to_query
        ]
        if (
            len(queries) == 0
        ):  # on a new database, no schemas are present and therefore, no objects exist (apart from public and information_schema)
            return []
        with ThreadPoolExecutor(max_workers=parallel_threads) as pool:
            results = pool.map(
                lambda q: self._snow_client.execute_query(q, use_dict_cursor=True),
                queries,
            )
        results_flat = [item for subresult in results for item in subresult]
        return results_flat

    def _get_constraint_metadata(
        self,
        schemas_to_query: List[str],
        parallel_threads,
        table_metadata_list: List[Dict],
    ) -> List[Dict]:
        """
        Get metadata for constraints and add them to given table metadata.
        This function exists because it is not possible to perform the required preprocessing of the data in Snowflake.
        (The information is only available in SHOW commands, which cannot be joined or filtered).
        """
        constraints_mapping = {
            "constraint_foreign_keys": {
                "template": os.path.join(
                    self._sql_files_path, "get_metadata_constraint_foreign_keys.sql"
                ),
                "schema_key": "fk_schema_name",
                "name_key": "fk_table_name",
            },
            "constraint_primary_keys": {
                "template": os.path.join(
                    self._sql_files_path, "get_metadata_constraint_primary_keys.sql"
                ),
                "schema_key": "schema_name",
                "name_key": "table_name",
            },
            "constraint_unique_keys": {
                "template": os.path.join(
                    self._sql_files_path, "get_metadata_constraint_unique_keys.sql"
                ),
                "schema_key": "schema_name",
                "name_key": "table_name",
            },
        }
        if len(table_metadata_list) == 0:
            return table_metadata_list
        database_name = table_metadata_list[0][
            "DATABASE_NAME"
        ]  # constraint SHOW commands require database name, should be the same for all entries in list
        for constraint, constraint_entry in constraints_mapping.items():
            metadata_constraints = [
                m
                for m in self._query_metadata(
                    constraint,
                    constraint_entry["template"],
                    schemas_to_query,
                    parallel_threads,
                    database_name,
                )
            ]
            for t in table_metadata_list:
                table_constraint_info = [
                    c
                    for c in metadata_constraints
                    if (
                        c[constraint_entry["schema_key"]] == t["SCHEMA_NAME"]
                        and c[constraint_entry["name_key"]] == t["OBJECT_NAME"]
                    )
                ]
                t.update({constraint: table_constraint_info})
        return table_metadata_list
    

    def _extend_metadata_policy_references(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        Add information on which row access policy is referenced where.
        """
        query_template = file_util.load(
            os.path.join(self._sql_files_path, "get_policy_references.sql")
        )
        for metadata_entry in metadata_list:
            if metadata_entry["kind"] not in ("MASKING_POLICY", "ROW_ACCESS_POLICY"):
                raise ValueError("Policy references can only be returned for masking policy or row access policy")
            query = query_template.format(
                schema_name=metadata_entry["schema_name"],
                object_name=metadata_entry["name"],
            )
            result = self._snow_client.execute_query(query, use_dict_cursor=True) # TODO: consider parallelization
            metadata_entry["references"] = result
        return metadata_list
            

    def _query_policies_on_objects_legacy(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        Query policies currently applied to objects in the given metadata list.
        """
        query_template = file_util.load(
            os.path.join(self._sql_files_path, "get_policies_on_single_object.sql")
        )
        for metadata_entry in metadata_list:
            if metadata_entry["TABLE_TYPE"] == "VIEW":
                query = query_template.format(
                    schema_name=metadata_entry["SCHEMA_NAME"],
                    object_name=metadata_entry["OBJECT_NAME"],
                    object_type=metadata_entry["TABLE_TYPE"],
                )
                result = self._snow_client.execute_query(query, use_dict_cursor=True)
                metadata_entry["APPLIED_POLICIES_LEGACY"] = result
            else:
                raise NotImplementedError(
                    "Getting policies for table type [ '{metadata_entry['TABLE_TYPE']}' ] not implemented."
                )
        return metadata_list

    def _extend_metadata_maskingpolicies(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        Policy metadata needs to be collected for each object individually, since snowflake does not provide any other option.
        """
        query_template = file_util.load(
            os.path.join(self._sql_files_path, "get_maskingpolicy_metadata.sql")
        )
        return self._extend_metadata_policies(metadata_list, query_template)

    def _extend_metadata_rowaccesspolicies(
        self, metadata_list: List[Dict]
    ) -> List[Dict]:
        """
        Policy metadata needs to be collected for each object individually, since snowflake does not provide any other option.
        """
        query_template = file_util.load(
            os.path.join(self._sql_files_path, "get_rowaccesspolicy_metadata.sql")
        )
        return self._extend_metadata_policies(metadata_list, query_template)

    def _extend_metadata_policies(
        self, metadata_list: List[Dict], query_template: str
    ) -> List[Dict]:
        """
        Policy metadata needs to be collected for each object individually, since snowflake does not provide any other option.
        """
        for metadata_entry in metadata_list:
            query = query_template.format(
                schema_name=metadata_entry["schema_name"],
                policy_name=metadata_entry["name"],
            )
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            metadata_entry.update(result[0])
        return metadata_list

    def _extend_metadata_pipes(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        INFORMATION_SCHEMA.PIPES does not contain information on INTEGRATION. We get this information here.
        #TODO: Evaluate if we should use SHOW PIPES IN SCHEMA instead of querying each pipe with DESC individually. (Information is the same in both cases.)
        """
        query_template = 'DESC PIPE "{schema_name}"."{pipe_name}";'
        for metadata_entry in metadata_list:
            query = query_template.format(
                schema_name=metadata_entry["SCHEMA_NAME"],
                pipe_name=metadata_entry["OBJECT_NAME"],
            )
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            pipe_status = self._get_pipe_status(
                metadata_entry["SCHEMA_NAME"], metadata_entry["OBJECT_NAME"]
            )
            execution_state = {"execution_state": pipe_status["executionState"]}
            result[0].update(execution_state)
            metadata_entry.update(result[0])
        return metadata_list

    def _get_pipe_status(self, pipe_schema: str, pipe_name: str) -> Dict:
        """
        Get the status of a pipe.
        """
        query = (
            f'SELECT SYSTEM$PIPE_STATUS(\'"{pipe_schema}"."{pipe_name}"\') AS STATUS'
        )
        result = self._snow_client.execute_query(query, use_dict_cursor=True)
        return json.loads(result[0]["STATUS"])

    def _extend_metadata_stages(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        INFORMATION_SCHEMA.STAGES and SHOW STAGES do not contain all information. We get this information here.
        """
        parent_properties = [
            "STAGE_FILE_FORMAT",
            "STAGE_COPY_OPTIONS",
            "STAGE_LOCATION",
            "STAGE_INTEGRATION",
            "DIRECTORY",
        ]
        query_template = 'DESC STAGE "{schema_name}"."{stage_name}";'
        for metadata_entry in metadata_list:
            query = query_template.format(
                schema_name=metadata_entry["schema_name"],
                stage_name=metadata_entry["name"],
            )
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            # result has the form of an array with dicts: [{parent_property, property, property_value}, {...}, ...]
            # convert this into a nested dict: {parent_property1: {property1:property_value1, property2:property_value2, ...}, ...}
            for parent_property in parent_properties:
                properties = [
                    r["property"]
                    for r in result
                    if r["parent_property"] == parent_property
                ]
                parent_property_value = {
                    r["property"]: r["property_value"]
                    for r in result
                    if (
                        r["parent_property"] == parent_property
                        and r["property"] in properties
                    )
                }
                # set PARSE_HEADER parameter on the default value ('false') if the key "PARSE_HEADER" does not exist - handles outdated instances of stages for which the parameter is missing
                if parent_property == "STAGE_FILE_FORMAT" and not "PARSE_HEADER" in parent_property_value:
                    parent_property_value["PARSE_HEADER"]='false'
                
                # The MULTI_LINE property of STAGE_FILE_FORMAT was introduced by the Snowflake Change Bundle 2024_08 and will be ignored because it can not be set or altered (yet)
                if parent_property == "STAGE_FILE_FORMAT":
                    parent_property_value.pop("MULTI_LINE", None)

                metadata_entry[parent_property] = parent_property_value
            
        return metadata_list

    def _get_meta_data_schema_names(self) -> List[str]:
        """
            Gets the schema names on the current database.
        Returns:
            List[str] - List of schema names
        """
        log.debug("GET list of SCHEMAS")
        query = "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME NOT IN ('INFORMATION_SCHEMA');"
        result = self._snow_client.execute_query(query)
        return [s["SCHEMA_NAME"] for s in result]

    def _get_filtered_schema_name_list(
        self, schema_list: Dict[str, List[str]]
    ) -> List[str]:
        """
        Get a list of all schemas on the database. Filter by schema_list.

        schema_list is a dictionary of either 'whitelist' or 'blacklist'.
        """
        schemas_on_db = self._get_meta_data_schema_names()
        if "whitelist" in schema_list:
            log.debug(
                f"FILTER list of schemas with WHITELIST [ {schema_list['whitelist']} ]"
            )
            schemas_whitelist = [s.casefold() for s in schema_list["whitelist"]]
            schemas_filtered = [
                s for s in schemas_on_db if s.casefold() in schemas_whitelist
            ]
        elif "blacklist" in schema_list:
            log.debug(
                f"FILTER list of schemas with BLACKLIST [ {schema_list['blacklist']} ]"
            )
            schemas_blacklist = [s.casefold() for s in schema_list["blacklist"]]
            schemas_filtered = [
                s for s in schemas_on_db if s.casefold() not in schemas_blacklist
            ]
        else:
            raise EnvironmentError(
                "MALFORMED config value [ 'schema_list' ] (contains neither blacklist nor whitelist)"
            )
        return schemas_filtered

    def get_all_objects(self, database_name: str) -> List[List[Union[str, int]]]:
        """
            Gets a list of all tables, views, functions, procedures, file formats and stages on a Snowflake database.
        Args:
            database_name: str - Name of the database
        Returns:
             List[List[Union[str,int]]] - Raw query return set provided by Snowflake
        """
        query = file_util.load(
            os.path.join(self._sql_files_path, "get_all_objects.sql")
        ).format(database_name=database_name)
        return self._snow_client.execute_query(query, use_dict_cursor=True)

    def get_all_objects_filtered(
        self, database_name: str, schema_list: List[str]
    ) -> List[List[Union[str, int]]]:
        """
            Gets a list of all tables, views, functions, procedures, file formats and stages on a Snowflake database,
            filtered by a list of given schemas.
        Args:
            database_name: str - Name of the database
            schema_list: List[str] - List of schemas to include
        Returns:
             List[List[Union[str,int]]] - Raw query return set provided by Snowflake
        """
        result = self.get_all_objects(database_name)
        return [
            r
            for r in result
            if r["SCHEMA_NAME"].lower() in [s.lower() for s in schema_list]
        ]

    def get_object_count_per_schema(self, database_name) -> List[Dict]:
        """
        Get a list of schemas in the database with a count of objects in each, ordered by number of objects (descending).

        Example output:
        [{'COUNT': 5, 'SCHEMA_NAME': 'S1'}, {'COUNT': 3, 'SCHEMA_NAME': 'S2'}]
        """
        query = file_util.load(
            os.path.join(self._sql_files_path, "get_object_count_per_schema.sql")
        ).format(database_name=database_name)
        return self._snow_client.execute_query(query, use_dict_cursor=True)
    

    @staticmethod
    def _convert_column_details_to_dict(metadata_list: List[dict]) -> None:
        """
        Given a metadata list for a table like object, convert the COLUMN_DETAILS from a json string to a list of dicts.
        """
        for metadata_object in metadata_list:
            if "COLUMN_DETAILS" in metadata_object.keys():
                metadata_object["COLUMN_DETAILS"] = json.loads(metadata_object["COLUMN_DETAILS"])
    
    def _extend_metadata_with_tags(        
            self, metadata_list: List[Dict], max_number_of_threads: int, object_type: str
    ) -> List[Dict]:
        """
        Get all tags on an table-like object (e.g. tables, views, materialized views).
        Here, each object is queried individually.
        """
        start_time_extend_metadata_with_tags = time.time()
        object_domain_tags = DbObjectType.get_object_domain_from_object_type_for_tag_references(object_type)

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_with_tags=self._get_metadata_tags_on_objects(object_domain_tags, metadata_list)
        else:
            # with threading
            metadata_list_extended_with_tags=parallelization_util.execute_func_in_parallel(self._get_metadata_tags_on_objects, metadata_list, max_number_of_threads, object_domain_tags)

        end_time_extend_metadata_with_tags = time.time()
        log.debug(f"EXECUTED _extend_metadata_with_tags for object_type {object_type}: [ '{round(end_time_extend_metadata_with_tags - start_time_extend_metadata_with_tags, 2)} seconds' ]")

        return metadata_list_extended_with_tags
    
    def _get_metadata_tags_on_objects(self, object_domain_tags: str, metadata_list: list[Dict])-> list[Dict]:
        """
        Get all tags on a list of objects and add them to the metadata information.
        """

        for metadata_entry in metadata_list:
            metadata_entry_database_name=metadata_entry["DATABASE_NAME"]
            metadata_entry_schema_name=metadata_entry["SCHEMA_NAME"]
            metadata_entry_object_name=metadata_entry["OBJECT_NAME"]
            
            metadata_entry["tags"]={}
           
            if object_domain_tags == 'TABLE':
                metadata_entry_column_tags={}
                query = f"SELECT * FROM TABLE(SNOWFLAKE.INFORMATION_SCHEMA.TAG_REFERENCES_ALL_COLUMNS('{metadata_entry_database_name}.{metadata_entry_schema_name}.{metadata_entry_object_name}', '{object_domain_tags}'));"
                result = self._snow_client.execute_query(query, use_dict_cursor=True)
                for row in result: 
                    if row["LEVEL"]=='TABLE':
                        metadata_entry["tags"][f'{row["TAG_SCHEMA"]}.{row["TAG_NAME"]}']=row["TAG_VALUE"]
                    elif row["LEVEL"]=='COLUMN':
                        if row["COLUMN_NAME"] not in metadata_entry_column_tags.keys():
                            metadata_entry_column_tags[row["COLUMN_NAME"]]={f'{row["TAG_SCHEMA"]}.{row["TAG_NAME"]}':row["TAG_VALUE"]}
                        else:
                            metadata_entry_column_tags[row["COLUMN_NAME"]].update({f'{row["TAG_SCHEMA"]}.{row["TAG_NAME"]}':row["TAG_VALUE"]})
                    
                for column_detail in metadata_entry["COLUMN_DETAILS"]:
                    if column_detail["COLUMN_NAME"] in metadata_entry_column_tags.keys():
                        column_detail["tags"]=metadata_entry_column_tags[column_detail["COLUMN_NAME"]]
                    else:
                        column_detail["tags"]={}

        return metadata_list




import logging
import os
import re

import git
from acedeploy.core.model_git_entities import GitFile
from aceutils.logger import LoggingAdapter, LogOperation, LogStatus

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class GitClient(object):
    """
    Client for any operation regarding the underlying git repository
    """

    def __init__(
        self, solution_root_path: str, git_tag_regex: str, deployment_mode: str
    ) -> None:
        """
            Inits a new git client.
        Args:
            solution_root_path: str - root path of git repo
            git_tag_regex: str - regular expression by which to find the starting commit
            deployment_mode: str - 'validate' or any other string
        """
        self.git_tag_regex = git_tag_regex
        self.deployment_mode = deployment_mode
        self.solution_root_path = solution_root_path
        self.repository = git.Repo(self.solution_root_path)
        self.commit_start = None
        self.commit_end = None

    def get_diff_filelist(self):
        """
        Evaluates the changes found in the solution git repository
        and returns a collection of changes with meta information.

        Will take latest commit that matches git_tag_regex as first commit.
        Will take current commit as last commit.

        For influence of deployment_mode, see doc/deployment_settings.md
        """
        # region get commit start and commit end information
        tags_sorted = sorted(
            self.repository.tags, key=lambda t: t.commit.committed_datetime
        )
        commit_start = None
        self.commit_end = self.repository.commit("HEAD")
        for tag in reversed(tags_sorted):
            if re.search(self.git_tag_regex, tag.name):
                commit_start = tag.commit
                break
        if commit_start is None:
            raise ValueError(
                f"ERROR: GIT mode REQUIRES at least one tag WITH format [ '{self.git_tag_regex}' ]"
            )
        else:
            self.commit_start = commit_start
        # endregion get commit start and commit end information

        # region get changes
        if self.deployment_mode.casefold() == "validate":
            common_ancestor = self.repository.merge_base(
                self.commit_start, self.commit_end
            )
            log.info(
                f"FOUND start commit: {self.commit_start.name_rev} AND end commit: {self.commit_end.name_rev} WITH common ancestor(s) {[c.name_rev for c in common_ancestor]}",
                operation=LogOperation.GIT,
                status=LogStatus.SUCCESS,
            )
            log.info(
                f"USE diff from {[c.name_rev for c in common_ancestor]} to {self.commit_end.name_rev} (equivalent to 'git diff {self.commit_start.hexsha}...{self.commit_end.hexsha}')"
            )
            diffs = self.commit_end.diff(common_ancestor, R=True)
        else:
            log.info(
                f"FOUND start commit: {self.commit_start.name_rev} AND end commit: {self.commit_end.name_rev}",
                operation=LogOperation.GIT,
                status=LogStatus.SUCCESS,
            )
            log.info(
                f"USE diff from {self.commit_start.name_rev} to {self.commit_end.name_rev} (equivalent to 'git diff {self.commit_start.hexsha}..{self.commit_end.hexsha}')"
            )
            diffs = self.commit_start.diff(self.commit_end)

        git_changed_files_list = [
            GitFile(
                os.path.join(
                    self.solution_root_path, diff.b_rawpath.decode(encoding="UTF-8")
                ),
                diff.change_type,
            )
            for diff in diffs
        ]
        # endregion get changes

        log.info(
            f"SUMMARY git: ADDED {len([f for f in git_changed_files_list if f.change_type == 'A'])} file(s), MODIFIED {len([f for f in git_changed_files_list if f.change_type in ('M', 'R')])} file(s), DELETED {len([f for f in git_changed_files_list if f.change_type == 'D'])} file(s)",
            operation=LogOperation.GIT,
            status=LogStatus.SUCCESS,
        )

        for f in git_changed_files_list:
            if f.change_type == "A":
                change_verb = "ADDED"
            elif f.change_type in ("M", "R"):
                change_verb = "MODIFIED"
            elif f.change_type == "D":
                change_verb = "DELETED"
            log.debug(f"{change_verb} file [ '{f.file_name}' ]")

        return git_changed_files_list



import logging

from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
from acedeploy.core.model_instance_objects import InstanceObject
from acedeploy.services.metadata_service import MetadataService
from acedeploy.services.policy_service import PolicyAssignmentObjectTypes, PolicyAssignmentAlterObjectTypes, PolicyHandlingTypes, PolicyService
from acedeploy.services.solution_service import SolutionClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class DbCompareClient(object):
    """
    Compare two databases states (current state and desired state).
    Generate a list of actions to change current state into desired state.
    Requires current state to be generated based on solution.
    All objects of desired state need to be DDLs in solution.
    """

    def __init__(
        self,
        desired_state: MetadataService,
        current_state: MetadataService,
        solution_client: SolutionClient,
    ) -> None:
        """
            Inits a new DbCompareClient
        Args:
            desired_state: MetadataService - Database state that should be achieved
            current_state: MetadataService - Current database state
            solution_client: SolutionClient - SQL DDL solution on disk
        """
        self._desired_state = desired_state
        self._current_state = current_state
        self._solution_client = solution_client
        self.action_list = []
        self.policy_assignments_handler = {}
        self._check_desired_state_in_solution()

    def _check_desired_state_in_solution(self):
        """
        Sanity check: Is every item in _desired_state in solution_client?
        The normal workflow is to create the desired_state through a deployment
        of (a subset of) the SQL solution to a database.
        If there are items in _desired_state that are not in the solution,
        something went wrong somewhere.
        """
        log.debug("CHECK all desired state objects in sql solution")
        for obj in self._desired_state.all_objects:
            if not self._solution_client.get_object_by_object(obj):
                raise ValueError(
                    f"The database object [ '{obj}' ] appears in _desired_state, but not in _solution_client."
                )
    
    def _fill_policy_assignments_handler(self, current_instance: InstanceObject , desired_instance: InstanceObject, object_type: str, policy_handling_type: PolicyHandlingTypes):
        """
        Fill policy assignments handler with current instance and desired instance of an object for which the policy assignments needs to be checked/handled.
        """
        current_instance_name, current_instance_columns = PolicyService.parse_instance_for_policy_assignments_handler(current_instance)

        desired_instance_name, desired_instance_columns = PolicyService.parse_instance_for_policy_assignments_handler(desired_instance)

        if current_instance_name:
            instance_name=current_instance_name
        else:
            instance_name=desired_instance_name

        if not (policy_handling_type == PolicyHandlingTypes.ADD_TABLECOLUMN and set(desired_instance_columns) == set(current_instance_columns)):

            if object_type not in self.policy_assignments_handler:
                self.policy_assignments_handler[object_type]= {}

            if current_instance:
                self.policy_assignments_handler[object_type][instance_name]={"policy_handling_type": policy_handling_type.value, "current_instance":{"columns": current_instance_columns}, "desired_instance":{"columns": desired_instance_columns}}
            else:
                self.policy_assignments_handler[object_type][instance_name]={"policy_handling_type": policy_handling_type.value, "desired_instance":{"columns": desired_instance_columns}}

            
    def get_add_actions(self) -> None:
        """
        Get all ADD actions and add them to action list.
        Add all objects that are in _desired_state, but not in _current_state.
        """
        log.debug("ADD actions of type [ 'ADD' ] to action_list")
        for obj in self._desired_state.all_objects:
            current_obj = self._current_state.get_object_by_object(obj)
            if not current_obj:
                log.debug(f"ADD action of type [ 'ADD' ] for object [ '{str(obj)}' ]")
                solution_object = self._solution_client.get_object_by_object(obj)
                self.action_list.append(
                    DbObjectAction.factory_from_solution_object(
                        solution_object, DbActionType.ADD, desired_instance=obj
                    )
                )
                if obj.object_type.name in PolicyAssignmentObjectTypes._member_names_:
                    self._fill_policy_assignments_handler(
                        current_instance={},
                        desired_instance=obj,
                        object_type=obj.object_type.name,
                        policy_handling_type=PolicyHandlingTypes.NEW_OBJECT
                    )

    def get_drop_actions_from_solution(self) -> None:
        """
        Based on the solution, get all DROP actions and add them to the action_list.
        Add all objects which are in _current_state, but not in the SQL solution.
        """
        log.debug(
            "ADD actions of type [ 'DROP' ] to action_list using information from SOLUTION"
        )
        for obj in self._current_state.all_objects:
            solution_obj = self._solution_client.get_object_by_object(obj)
            if not solution_obj:
                log.debug(f"ADD action of type [ 'DROP' ] for object [ '{str(obj)}' ]")
                self.action_list.append(
                    DbObjectAction.factory_from_instance_object(obj, DbActionType.DROP)
                )

    def get_drop_overloaded_actions(self, drop_overloaded_procedures: bool, drop_overloaded_functions: bool) -> None:
        """
        Get DROPOVERLOADED actions for overloaded objects (e.g. procedures, functions) and add them to the action_list.
        This is required if the user has configured to drop overloaded objects.
        """
        log.debug("ADD actions of type [ 'DROP' ] for overloaded objects to action_list")
        for obj in self._desired_state.all_objects:
            if (
                obj.object_type == DbObjectType.PROCEDURE 
                and drop_overloaded_procedures 
                and obj.id in [action.id for action in self.action_list if action.object_type == DbObjectType.PROCEDURE] # check if the procedure is in the action list to only drop overloaded procedures for which we have an ADD or ALTER action
            ):
                overloaded_procedures = [proc for proc in self._current_state.procedures if proc.name == obj.name and proc.schema == obj.schema]
                for overloaded_proc in overloaded_procedures:
                    log.debug(f"ADD action of type [ 'DROPOVERLOADED' ] for object [ '{str(overloaded_proc)}' ]")
                    self.action_list.append(
                        DbObjectAction.factory_from_instance_object(overloaded_proc, DbActionType.DROPOVERLOADED)
                    )
            elif (
                obj.object_type == DbObjectType.FUNCTION 
                and drop_overloaded_functions
                and obj.id in [action.id for action in self.action_list if action.object_type == DbObjectType.FUNCTION] # check if the function is in the action list to only drop overloaded functions for which we have an ADD or ALTER action
            ):
                overloaded_functions = [func for func in self._current_state.procedures if func.name == obj.name and func.schema == obj.schema]
                for overloaded_func in overloaded_functions:
                    log.debug(f"ADD action of type [ 'DROPOVERLOADED' ] for object [ '{str(overloaded_func)}' ]")
                    self.action_list.append(
                        DbObjectAction.factory_from_instance_object(overloaded_func, DbActionType.DROPOVERLOADED)
                    )

    def get_drop_actions_from_git(self) -> None:
        """
        Based on the git diff, get all DROP actions and add them to the action_list.
        Add all objects which have git change type DELETED and appear in _current_state.
        This function will not add objects which have been RENAMED in git and therefore no longer appear in the solution.
        """
        raise NotImplementedError("Drop actions based on git diff is not implemented.")
        # CODE BELOW DOES NOT WORK: deleted objects do not appear in _solution_client
        # log.debug(f"ADD actions of type [ 'DROP' ] to action_list using information from GIT")
        # for solution_obj in self._solution_client.all_objects:
        #     if solution_obj.git_change_type == 'D': # this does not work: deleted objects do not appear in _solution_client
        #         current_obj = self._current_state.get_object_by_object(solution_obj)
        #         if current_obj:
        #             self._add_to_action_list(current_obj, DbActionType.DROP)

    def get_alter_actions(self) -> None:
        """
        Get ALTER actions and add them to the action_list.
        Go through all objects which appear both in _current_state and in _desired_state.
            - for objects that we will alter via ALTER statement: add to action_list, if difference between boths states detected
            - for all other object (which we will alter via CREATE OR REPLACE statement): add this object to action list.
        """
        log.debug("ADD actions of type [ 'ALTER' ] to action_list")
        for desired_obj in self._desired_state.all_objects:
            current_obj = self._current_state.get_object_by_object(desired_obj)
            if current_obj:
                if current_obj.object_type in (
                    DbObjectType.SCHEMA,
                    DbObjectType.TABLE,
                    DbObjectType.VIEW,
                    DbObjectType.SEQUENCE,
                    DbObjectType.STREAM,
                    DbObjectType.MASKINGPOLICY,
                    DbObjectType.ROWACCESSPOLICY,
                    DbObjectType.PIPE,
                    DbObjectType.STAGE,
                    DbObjectType.FILEFORMAT,
                    DbObjectType.DYNAMICTABLE,
                    DbObjectType.HYBRIDTABLE,
                    DbObjectType.NETWORKRULE,
                    DbObjectType.TAG,
                    # DbObjectType.FUNCTION, # functions can currently not be added here: we do not query all required metadata (e.g. HANDLER, parameter default values)
                    # DbObjectType.PROCEDURE, # procedures can currently not be added here: we do not query all required metadata (e.g. EXECUTE AS, parameter default values)
                    # DbObjectType.MATERIALIZEDVIEW, # materialized views can currently not be added here: we do not query all required metadata (e.g. DEFINITION)
                    # DbObjectType.TASK, # with the current implementation, we can not add tasks here (i.e. conditionally if a change is detected). we currently rely on always adding the full tree if any task in that tree is changed. if we did not add the full tree, we would need to take care of properly suspending tasks in that tree that we do not want to alter.
                    # DbObjectType.MATERIALIZEDTABLE, # materialized tables can currently not be added here: we do not query all required metadata (e.g. AUTO REFRESH)
                ):
                    if current_obj != desired_obj:
                        log.debug(
                            f"ADD action of type [ 'ALTER' ] for object [ '{str(current_obj)}' ]"
                        )
                        solution_object = self._solution_client.get_object_by_object(
                            desired_obj
                        )

                        alter_action= DbObjectAction.factory_from_instance_objects(
                                current_obj,
                                desired_obj,
                                DbActionType.ALTER,
                                file_content=solution_object.content,
                            )

                        self.action_list.append(alter_action)

                        if current_obj.object_type.name in PolicyAssignmentAlterObjectTypes._member_names_:
                            policy_handling_type= PolicyAssignmentAlterObjectTypes.get_policy_handling_type(PolicyAssignmentAlterObjectTypes(current_obj.object_type.name))
                            self._fill_policy_assignments_handler(
                                current_instance=current_obj,
                                desired_instance=desired_obj,
                                object_type=current_obj.object_type.name,
                                policy_handling_type=policy_handling_type
                            )
                else:
                    log.debug(
                        f"ADD action of type [ 'ALTER' ] for object [ '{str(desired_obj)}' ]"
                    )
                    solution_object = self._solution_client.get_object_by_object(
                        desired_obj
                    )
                    if current_obj.object_type == DbObjectType.TASK:
                        self.action_list.append(
                            DbObjectAction.factory_from_solution_object(
                                solution_object,
                                DbActionType.ALTER,
                                state_before_deployment=current_obj.state,
                            )
                        )
                    else:
                        self.action_list.append(
                            DbObjectAction.factory_from_solution_object(
                                solution_object, DbActionType.ALTER
                            )
                        )



import logging
import re
import inspect
from pathlib import PurePath
from collections import defaultdict

import aceutils.file_util as file_util
from aceaccount.configuration import AccountObjectConfig
from aceaccount.services.account_objects_solution_service import AccountSolutionClient
from aceservices.snowflake_service import SnowClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class AccountValidationService:
    """
    Class to validate account objects on a technical level and on a content level.
    """

    def __init__(self, config: AccountObjectConfig, snow_client: SnowClient = None):
        self._config = config
        self._snow_client = snow_client

    def validate_against_json_schema(
        self, jsonschema_absolute_folder_path: str, jsonschema_file_name: str
    ):
        """
        Validate account objects against a JSON schema.
        Raises an Error when the validation fails.
        """
        try:
            account_objects_jsonschema_path_content = PurePath(
                jsonschema_absolute_folder_path
            ).joinpath(PurePath(jsonschema_file_name))

            for source_file in self._config.source_files:
                if (
                    file_util.validate_json(
                        account_objects_jsonschema_path_content,
                        source_file,
                        jsonschema_absolute_folder_path,
                    )
                    is False
                ):
                    raise EnvironmentError(
                        f"FAILED validation of {source_file} \n against schema {account_objects_jsonschema_path_content}"
                    )
                else:
                    log.info(
                        "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
                    )
                    log.info(
                        f"SUCCEEDED validation of {source_file} \n  against schema {account_objects_jsonschema_path_content}"
                    )

        except Exception as err:
            log.error(str(err))
            raise err

    def validate_tags(self, solution_client: AccountSolutionClient):
        tags = set()
        tag_values_per_tag = defaultdict(list)
        for account_object in solution_client.all_account_objects:
            if hasattr(account_object, "tags"):
                tags.update(account_object.tags.keys())
                for tag, tag_value in account_object.tags.items():
                    tag_values_per_tag[tag].append(tag_value)

        tags = list(tags)

        self._validate_tags_check_fully_qualified_identifiers(tags)
        self._validate_tag_values_against_allowed_values(tag_values_per_tag)

    def _validate_tags_check_fully_qualified_identifiers(self, tags: list):
        for tag in tags:
            if not len(tag.split(".")) == 3:
                raise ValueError(
                    f"The Snowflake tag {tag} is not fully identified with Snowflake database and Snowflake schema."
                )

    def _validate_tag_values_against_allowed_values(self, tag_values_per_tag: dict):
        for tag, tag_values in tag_values_per_tag.items():
            query = f"SELECT SYSTEM$GET_TAG_ALLOWED_VALUES('{tag}');"
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            allowed_tag_values_string = list(result[0].values())[0]
            if allowed_tag_values_string:
                allowed_tag_values = allowed_tag_values_string[2:-2].split('","')
                for tag_value in tag_values:
                    if tag_value not in allowed_tag_values:
                        raise ValueError(
                            f"The tag value {tag_value} is not in the allowed tag values for the tag {tag} as defined on the Snowflake Account."
                        )

def validate_additionally_storage_integration_allowed_locations(config: AccountObjectConfig, additional_validation_folder_path: str='', additional_validation_file_name: str=''):
    """
    Validate the storage_allowed_locations parameter of storage integrations.
    Check if the name of the integration exists in all paths of the storage_allowed_locations.
    Raises an Error when the validation fails.
    """
    storage_allowed_locations_not_in_line = {}
    additional_validation_config = {}
    storage_integration_exceptions = []
    validated_source_files = []

    if additional_validation_folder_path and additional_validation_file_name:
        additional_validation_file_path = PurePath(
                    additional_validation_folder_path
                ).joinpath(PurePath(additional_validation_file_name))

        additional_validation_config=file_util.load_json(additional_validation_file_path)
    
    for source_file in config.source_files:
        account_object_solution = file_util.load_json(source_file)
        if "storage_integrations" not in account_object_solution.keys():
            continue
        else:
            validated_source_files.append(source_file)
            for storage_integration_name, storage_integration_parameters  in account_object_solution["storage_integrations"].items():

                storage_integration_name_split=storage_integration_name.lower().split('sti_', 1)
                if len(storage_integration_name_split) == 1:
                    storage_integration_identifier=storage_integration_name_split[0]
                else:
                    storage_integration_identifier=storage_integration_name_split[1]   

                if additional_validation_config:
                    storage_integration_exceptions=additional_validation_config["exceptions_storage_integration_allowed_locations"].get(storage_integration_name)

                if storage_integration_exceptions:
                    pattern = rf'^azure:\/\/(.*{storage_integration_identifier}.*|{"|".join(storage_integration_exceptions)})(.blob.core.windows.net\/).+|^[*]{1}$'
                else:
                    pattern = rf'^azure:\/\/.*{storage_integration_identifier}.*(.blob.core.windows.net\/).+|^[*]{1}$'
                
                if not (storage_integration_exceptions and "*" in storage_integration_exceptions):
                    for storage_allowed_location in storage_integration_parameters["storage_allowed_locations"]:
                        m = re.match(pattern, storage_allowed_location, re.IGNORECASE)
                        if not m: 
                            if storage_integration_name not in storage_allowed_locations_not_in_line:
                                storage_allowed_locations_not_in_line[storage_integration_name]=[]
                            storage_allowed_locations_not_in_line[storage_integration_name].append(storage_allowed_location)

    if not storage_allowed_locations_not_in_line:
        log.info(
            f"SUCCEEDED additional validation of storage_allowed_locations for storage integrations in the following source files {validated_source_files}."
        )
    else:
        raise ValueError(inspect.cleandoc(f"FAILED additional validation of storage_allowed_locations for storage integrations.\n\
                                            The following path/paths in the storage_allowed_locations is/are not allowed - the path/paths does/do not contain the name of the storage integration:\n\
                                            {storage_allowed_locations_not_in_line}"
                                        )
                        )



import logging
from typing import Dict, List

from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
from aceservices.snowflake_service import SnowClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def set_persist_tags(
    persist_table_name: str,
    action_list: List[DbObjectAction],
    database_name: str,
    snow_client: SnowClient,
):
    """
        Set the tag CREATE_FG for each altered and added view in action_list
    Params:
        persist_table_name: Fully qualified name of the persist table (database.schema.tablename)
        action_list: List of object actions that were performed during the deployment
        database_name: Name of the database to which the objects were deployed
        snow_client: account-scoped snowflake client
    """
    views = get_views(action_list)

    if len(views) == 0:
        log.info(
            f"SET tag CREATE_FG in table [ '{persist_table_name}' ] NOT required, no views deployed"
        )
        return

    statement_template = """
        UPDATE {persist_table_name}
        SET CREATE_FG = 1
        WHERE TARGET_DATABASE ILIKE '{database}'
        AND (UPPER(ORIGIN_TABLE_SCHEMA), UPPER(TABLE_NAME)) IN
        (
            {table_selects}
        );
    """
    table_select_template = "SELECT '{schema}', '{view}'"
    table_selects = []
    for view in views:
        table_selects.append(
            table_select_template.format(
                schema=view["schema_name"].upper(), view=view["view_name"].upper()
            )
        )

    statement = statement_template.format(
        persist_table_name=persist_table_name,
        database=database_name,
        table_selects="\nUNION ".join(table_selects),
    )

    log.info(
        f"SET tag CREATE_FG in table [ '{persist_table_name}' ] for [ '{len(table_selects)}' ] views"
    )
    snow_client.execute_statement(statement)


def get_views(action_list: List[DbActionType]) -> List[Dict[str, str]]:
    """
    Return a list of views that were added or altered
    """
    return [
        {"schema_name": a.schema, "view_name": a.name}
        for a in action_list
        if (
            (a.object_type == DbObjectType.VIEW)
            and (a.action in (DbActionType.ADD, DbActionType.ALTER))
        )
    ]



import logging
import os
import json
from pathlib import PurePath
from jsonschema import validate, RefResolver
import datetime
import time

from dataclasses import dataclass
import aceutils.file_util as file_util
from aceaccount.configuration import AccountObjectConfig
from aceservices.snowflake_service import SnowClient, SnowClientConfig
from aceaccount.services.account_objects_metadata_service import AccountMetadataService
from aceaccount.services.account_objects_solution_service import AccountSolutionClient
from aceaccount.services.account_objects_validation_service import AccountValidationService
from aceutils.logger import LoggingAdapter
from aceaccount.services.account_objects_compare_service import (
    AccountObjectCompareClient,
)

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)

@dataclass
class AccountObjectParams:
    technical_jsonschema_folder_path:str        = 'resources/json-schemas/account_object_json_schemas/technical_validation/'
    technical_jsonschema_file_name:str          = 'account_objects_technical_validation.schema.json'

def execute(
        config: AccountObjectConfig,
        dryrun: bool = False, 
        output_sql_statements: bool = False, output_path: str = '', 
        optional_technical_jsonschema_absolute_folder_path: str = None,
        optional_technical_jsonschema_file_name: str = None
        ):
    try:
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M_%S")

        start_time_validate_technical = time.time()
        validate_technical(config, optional_technical_jsonschema_absolute_folder_path, optional_technical_jsonschema_file_name)
        end_time_validate_technical = time.time()
        log.info(f"============= Execution Time validate_technical: {round(end_time_validate_technical - start_time_validate_technical, 2)} seconds")

        start_time_solution_client = time.time()
        solution_client = AccountSolutionClient(config.enabled_object_types, config.source_files)
        end_time_solution_client = time.time()
        log.info(f"============= Execution Time solution_client: {round(end_time_solution_client - start_time_solution_client, 2)} seconds")


        start_time_snow_client = time.time()
        snow_client_config = SnowClientConfig(
            account=config.snow_account,
            user=config.snow_login,
            password=config.snow_password,
            warehouse=config.snow_warehouse,
            role=config.snow_role,
        )

        snow_client = SnowClient(snow_client_config)
        end_time_snow_client = time.time()
        log.info(f"============= Execution Time snow_client: {round(end_time_snow_client - start_time_snow_client, 2)} seconds")

        start_time_metadata_client = time.time()
        metadata_client = AccountMetadataService(config.enabled_object_types, snow_client, dryrun)
        end_time_metadata_client = time.time()
        log.info(f"============= Execution Time metadata_client: {round(end_time_metadata_client - start_time_metadata_client, 2)} seconds")

        start_time_get_all_account_objects_metadata = time.time()
        metadata_client.get_all_account_objects_metadata(max_number_of_threads=config.max_number_of_threads)
        end_time_get_all_account_objects_metadata = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata: {round(end_time_get_all_account_objects_metadata - start_time_get_all_account_objects_metadata, 2)} seconds")

        start_time_generate_actions = time.time()
        compare_client = AccountObjectCompareClient(solution_client, metadata_client, snow_client_config)
        compare_client.generate_add_actions()
        if config.drop_enabled:
            compare_client.generate_drop_actions(config.drop_enabled)
        compare_client.generate_alter_actions()
        end_time_generate_actions = time.time()
        log.info(f"============= Execution Time generate_actions: {round(end_time_generate_actions - start_time_generate_actions, 2)} seconds")

        start_time_generate_statements = time.time()
        statements = []
        for action in compare_client.action_list:
            statement_object = action.generate_statement_object()
            statements.append(statement_object.statement)
        statements_text = '\n\n'.join(statements)
        if not statements_text:
            statements_text = ' '
        end_time_generate_statements = time.time()
        log.info(f"============= Execution Time generate_statements: {round(end_time_generate_statements - start_time_generate_statements, 2)} seconds")

        if not dryrun:
            start_time_execute_statements = time.time()
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("Executing the following statements - includes grant statements and set-tag statements for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            snow_client.execute_statement(statements)
            end_time_execute_statements = time.time()
            log.info(f"============= Execution Time execute_statements: {round(end_time_execute_statements - start_time_execute_statements, 2)} seconds")
        else:
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("The dry-run produced the following statements - includes grant statements and set-tag statements for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        
        start_time_generate_grant_actions = time.time()
        compare_client.generate_grant_actions()
        if config.revoke_enabled:
            compare_client.generate_revoke_actions()
        end_time_generate_grant_actions = time.time()
        log.info(f"============= Execution Time generate_grant_actions and generate_revoke_actions: {round(end_time_generate_grant_actions - start_time_generate_grant_actions, 2)} seconds")

        start_time_generate_grant_statements = time.time()
        account_objects_grant_statements = []
        for grant_action in compare_client.grant_action_list:
            grant_action_object = grant_action.generate_statement_object()
            account_objects_grant_statements.append(grant_action_object.statement)
        account_objects_grant_statements_text = '\n'.join(filter(None,account_objects_grant_statements))
        if not account_objects_grant_statements_text:
            account_objects_grant_statements_text = ' '
        end_time_generate_grant_statements = time.time()
        log.info(f"============= Execution Time generate_grant_statements: {round(end_time_generate_grant_statements - start_time_generate_grant_statements, 2)} seconds")

        if not dryrun:
            start_time_execute_grant_statements = time.time()
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("Executing the following grant statements for previously existing objects - does not include grants for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(account_objects_grant_statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            snow_client.execute_statement(account_objects_grant_statements)
            end_time_execute_grant_statements = time.time()
            log.info(f"============= Execution Time execute_grant_statements: {round(end_time_execute_grant_statements - start_time_execute_grant_statements, 2)} seconds")
        else:
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("The dry-run produced the following grant statements for previously existing objects - does not include grants for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(account_objects_grant_statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")

        start_time_generate_tag_actions = time.time()
        compare_client.generate_settag_actions()
        if config.unsettag_enabled:
            compare_client.generate_unsettag_actions()
        end_time_generate_tag_actions = time.time()
        log.info(f"============= Execution Time generate_settag_actions and generate_unsettag_actions: {round(end_time_generate_tag_actions - start_time_generate_tag_actions, 2)} seconds")

        start_time_generate_tag_statements = time.time()
        account_objects_tag_statements = []
        for tag_action in compare_client.tag_action_list:
            tag_action_object = tag_action.generate_statement_object()
            account_objects_tag_statements.append(tag_action_object.statement)
        account_objects_tag_statements_text = '\n'.join(filter(None,account_objects_tag_statements))
        if not account_objects_tag_statements_text:
            account_objects_tag_statements_text = ' '
        end_time_generate_tag_statements = time.time()
        log.info(f"============= Execution Time generate_tag_statements: {round(end_time_generate_tag_statements - start_time_generate_tag_statements, 2)} seconds")

        if not dryrun:
            start_time_execute_tag_statements = time.time()
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("Executing the following tag statements for previously existing objects - does not include tags for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(account_objects_tag_statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            snow_client.execute_statement(account_objects_tag_statements)
            end_time_execute_tag_statements = time.time()
            log.info(f"============= Execution Time execute_grant_statements: {round(end_time_execute_tag_statements - start_time_execute_tag_statements, 2)} seconds")
        else:
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info("The dry-run produced the following tag statements for previously existing objects - does not include tags for newly created objects:")
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            log.info(account_objects_tag_statements_text)
            log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        
        start_time_generate_output_sql_statements = time.time()
        if output_sql_statements:
            if output_path:
                filename_sql_statements = os.path.join(output_path, f"account_objects_sql_statements.{timestamp}.sql")
                filename_grant_statements = os.path.join(output_path, f"account_objects_grant_statements.{timestamp}.sql")
                filename_tag_statements = os.path.join(output_path, f"account_objects_tag_statements.{timestamp}.sql")
                log.info(f"Saving resulting sql statements of the account objects in file '{filename_sql_statements}' - includes grant statements for newly created objects.")
                log.info(f"Saving resulting grant statements of the account objects in file '{filename_grant_statements}' - does not include grants for newly created objects.")
                log.info(f"Saving resulting tag statements of the account objects in file '{filename_tag_statements}' - does not include tags for newly created objects.")
                os.makedirs(output_path, exist_ok=True)
                with open(filename_sql_statements, 'w') as f:
                    f.write(statements_text)
                with open(filename_grant_statements, 'w') as f:
                    f.write(account_objects_grant_statements_text)
                with open(filename_tag_statements, 'w') as f:
                    f.write(account_objects_tag_statements_text)
            else:
                log.info("No output_path defined. Statements are not saved in a file.")
        end_time_generate_output_sql_statements = time.time()
        log.info(f"============= Execution Time generate_output_sql_statements: {round(end_time_generate_output_sql_statements - start_time_generate_output_sql_statements, 2)} seconds")
        log.info("done")

    except Exception as err:
        log.error(str(err))
        raise err

def validate_technical(
        config, 
        optional_technical_jsonschema_absolute_folder_path: str = None,
        optional_technical_jsonschema_file_name: str = None
    ):
    """
        Technical validation of account objects.
        Raises an Error when the validation fails.
        Initializes the AccountSolutionClient as a test to check, e.g., for duplicates.

        Optional:
            Define optional JSON Schemas for the technical validation e.g. in the "setup-repository".
            This way, it is possible to circumvent the default restrictions by Snowflake in case there was a change by Snowflake just for a specific account.
            For example, if there is an exception on a Snowflake Account by which the max_cluster_count for a Warehouse can be set to 20 instead of the usual default maximum of 10.
            Use the following paramters to define the path to the optional JSON Schemas located outside of the acedeploy repo:
                optional_technical_jsonschema_absolute_folder_path
                optional_technical_jsonschema_file_name
            Note: Please define both of these parameters or none.
    """
    try:
        validation_service = AccountValidationService(config)

        module_root_folder=os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


        if optional_technical_jsonschema_absolute_folder_path and optional_technical_jsonschema_file_name:
            technical_jsonschema_absolute_folder_path = PurePath(module_root_folder).joinpath(PurePath(optional_technical_jsonschema_absolute_folder_path))
            technical_jsonschema_file_name = optional_technical_jsonschema_file_name
        else:
            technical_jsonschema_absolute_folder_path = PurePath(module_root_folder).joinpath(PurePath(AccountObjectParams.technical_jsonschema_folder_path))
            technical_jsonschema_absolute_folder_path = PurePath(module_root_folder).joinpath(PurePath(AccountObjectParams.technical_jsonschema_folder_path))
            technical_jsonschema_file_name = AccountObjectParams.technical_jsonschema_file_name
        
        validation_service.validate_against_json_schema(technical_jsonschema_absolute_folder_path, technical_jsonschema_file_name)
        
        # Test-Initialization of the AccountSolutionClient to check, e.g., for duplicates
        AccountSolutionClient(config.enabled_object_types, config.source_files)
    except Exception as err:
        log.error(str(err))
        raise err

def validate_content(config, content_jsonschema_absolute_folder_path: str, content_jsonschema_file_name: str):
    """
        Function to validate account objects in regards to the content.
        Raises an Error when the validation fails.
        Validates also the Snowflake tags in regards to fully-qulified identifiers, existence, and allowed values.
    """
    try:
        snow_client_config = SnowClientConfig(
            account=config.snow_account,
            user=config.snow_login,
            password=config.snow_password,
            warehouse=config.snow_warehouse,
            role=config.snow_role,
        )

        snow_client = SnowClient(snow_client_config)

        validation_service = AccountValidationService(config, snow_client)

        validation_service.validate_against_json_schema(content_jsonschema_absolute_folder_path, content_jsonschema_file_name)

        solution_client = AccountSolutionClient(config.enabled_object_types, config.source_files)
        
        start_time_validate_tags = time.time()
        validation_service.validate_tags(solution_client)
        end_time_validate_tags = time.time()
        log.info(f"============= Execution Time validate_tags: {round(end_time_validate_tags - start_time_validate_tags, 2)} seconds")
    except Exception as err:
        log.error(str(err))
        raise err
    
def fetch_account_objects(config: AccountObjectConfig, output_path: str, dialect_json_schema_relative_file_path: str, fetch_grants: bool=True, fetch_tags: bool=True):
    """
        Function to fetch existing account objects from Snowflake. 
        Generates a parameter representation and saves it as a JSON file which can be used as an input for an account objects pipeline.
        Raises an Error when the fetching fails.
    """
    try:
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M_%S")
        snow_client_config = SnowClientConfig(
                account=config.snow_account,
                user=config.snow_login,
                password=config.snow_password,
                warehouse=config.snow_warehouse,
                role=config.snow_role,
            )
        
        snow_client = SnowClient(snow_client_config)
        
        metadata_client = AccountMetadataService(config.enabled_object_types, snow_client)
        metadata_client.get_all_account_objects_metadata(max_number_of_threads=config.max_number_of_threads, get_grants=fetch_grants, get_tags=fetch_tags)
        fetched_account_objects=metadata_client.get_account_objects_dict(dialect_json_schema_relative_file_path)

        if output_path:
            os.makedirs(output_path, exist_ok=True)

            for object_type, account_objects_parameter_representation in fetched_account_objects.items():
                filename_account_objects = os.path.join(output_path, f"fetched_account_objects_of_type_{object_type}.{timestamp}.json")
                logging.info(f"SAVING fetched account objects as JSON file in '{filename_account_objects}'")
                with open(filename_account_objects, 'w') as f:
                    json.dump(account_objects_parameter_representation, f, indent=4)
        else:
            log.info("No output_path defined. Fetched account objects are not saved in a file.")
      
    except Exception as err:
        log.error(str(err))
        raise err


import logging
from typing import List
import copy
from collections import defaultdict

import aceutils.dict_and_list_util as dict_and_list_util
from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_solution_entities import SolutionObject
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
from acedeploy.services.dependency_parser import DependencyParser
from acedeploy.services.solution_service import SolutionClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def order_action_list(
    object_action_list: List[DbObjectAction],
    dependency_client: DependencyParser,
    is_meta_deployment: bool,
) -> List[List[List[DbObjectAction]]]:
    """
        Order a list of db object actions:
            - CREATE and ALTER schemas
            - CREATE and ALTER object (except tasks)
            - CREATE and ALTER tasks
            - DROP objects, except schemas
            (- currently disabled: DROP schemas)
    Args:
        object_action_list: List[DbObjectAction] - unordered list of db object actions
        solution_client: SolutionClient - contains all files loaded from disk
        is_meta_deployment: bool - if action list is used to deploy to a meta database
    Returns:
        ordered_object_action_list: List[List[List[DbObjectAction]]] - ordered list of db object actions
            - first list level: each list below needs to processed in order
            - second list level: this list contains collections steps that can be run in parallel
            - third level: this step collection must be executed in order
        Example (for clarity, using statements instead of DbObjectAction):
            [
                [
                    ["CREATE SCHEMA A"], ["CREATE SCHEMA B"]
                ], [
                    ["CREATE TABLE A.T1 ...", "CREATE VIEW A.V1 AS SELECT * FROM A.T1" ],
                    ["CREATE TABLE B.T2 ...", "CREATE VIEW A.V1 AS SELECT * FROM B.T2" ],
                    ["CREATE TABLE B.T3"]
                ], [
                    ["CREATE TASK A.T1 ..."], ["CREATE TASK B.T2 ..."]
                ], [
                    ["DROP VIEW C.V5"]
                ]
            ]
    """
    log.info(f"ORDER actions WITH count [ '{len(object_action_list)}' ]")
    ordered_object_action_list = []

    # always create new schemas first
    ordered_object_action_list.append(
        _generate_action_sublist_unordered(
            object_action_list,
            (DbObjectType.SCHEMA,),
            (DbActionType.ADD, DbActionType.ALTER),
        )
    )

    # always create new tags befor creating objects
    ordered_object_action_list.append(
        _generate_action_sublist_unordered(
            object_action_list,
            (DbObjectType.TAG,),
            (DbActionType.ADD, DbActionType.ALTER),
        )
    )

     # always drop overloaded procedures and functions before creating new ones (when Object-Option dropOverloadedObjects is set for procedures or functions)
    ordered_object_action_list.append(
        _generate_action_sublist_unordered(
            object_action_list,
            (DbObjectType.PROCEDURE, DbObjectType.FUNCTION ),
            (DbActionType.DROPOVERLOADED,),
        )
    )

    # add and alter objects
    ordered_object_action_list.append(
        _generate_action_sublist_ordered(
            object_action_list,
            (
                DbObjectType.VIEW,
                DbObjectType.MATERIALIZEDVIEW,
                DbObjectType.TABLE,
                DbObjectType.EXTERNALTABLE,
                DbObjectType.FUNCTION,
                DbObjectType.PROCEDURE,
                DbObjectType.FILEFORMAT,
                DbObjectType.SEQUENCE,
                DbObjectType.STAGE,
                DbObjectType.PIPE,
                DbObjectType.STREAM,
                DbObjectType.MASKINGPOLICY,
                DbObjectType.ROWACCESSPOLICY,
                DbObjectType.DYNAMICTABLE,
                DbObjectType.HYBRIDTABLE,
                DbObjectType.NETWORKRULE,
            ),
            (DbActionType.ADD, DbActionType.ALTER),
            dependency_client,
            is_meta_deployment,
        )
    )

    # add and alter tasks (task content is not validated on CREATE TASK, therefore they can by created independently of other objects)
    ordered_object_action_list.append(
        _generate_action_sublist_ordered(
            object_action_list,
            (DbObjectType.TASK,),
            (DbActionType.ADD, DbActionType.ALTER),
            dependency_client,
            is_meta_deployment,
        )
    )

    # objects to be dropped, except schemas
    ordered_object_action_list.append(
        _generate_action_sublist_unordered(
            object_action_list,
            [t for t in DbObjectType if t != DbObjectType.SCHEMA],
            (DbActionType.DROP,),
        )
    )

    # schemas to be dropped (currently disabled)
    # ordered_object_action_list.append(
    #     self._generate_action_sublist_unordered(
    #         object_action_list,
    #         (DbObjectType.SCHEMA,),
    #         (DbActionType.DROP,)
    #     )
    # )

    return ordered_object_action_list


def _generate_action_sublist_unordered(
    all_actions: List[DbObjectAction],
    object_types: List[DbObjectType],
    action_type: List[DbActionType],
) -> List[List[DbObjectAction]]:
    """
    Given a list of action, return all actions that match a specific DbObjectAction and DbActionType.

    It is assumed that the returned actions can be deployed without any interdependencies.
    Therefore, the each action will be returned in its own list. (Which can later be used for parallel deployment.)
    """
    return [
        [a]
        for a in all_actions
        if (a.object_type in object_types) and (a.action in action_type)
    ]


def _generate_action_sublist_ordered(
    all_actions: List[DbObjectAction],
    object_types: List[DbObjectType],
    action_type: List[DbActionType],
    dependency_client: DependencyParser,
    is_meta_deployment: bool,
) -> List[List[DbObjectAction]]:
    """
    Given a list of action, return all actions that match a specific DbObjectAction and DbActionType and order them.
    """
    object_action_list_to_be_ordered = [
        a
        for a in all_actions
        if (a.object_type in object_types) and (a.action in action_type)
    ]
    return _order_actions_with_internal_dependencies(
        object_action_list_to_be_ordered,
        dependency_client,
        is_meta_deployment,
    )


def order_actions_with_internal_dependencies_from_solution(
    object_types: List[DbObjectType],
    object_action_list_to_be_ordered: List[DbObjectAction],
    solution_client: SolutionClient,
    is_meta_deployment: bool,
) -> List[List[DbObjectAction]]:
    """
    Order a list of object actions with internal dependencies.
    """
    dep_client = DependencyParser(solution_client)
    dep_client.build_full_dependency_graph(object_types)
    return _order_actions_with_internal_dependencies(
        object_action_list_to_be_ordered, dep_client, is_meta_deployment
    )


def _order_actions_with_internal_dependencies(
    object_action_list_to_be_ordered: List[DbObjectAction],
    dependency_client: DependencyParser,
    is_meta_deployment: bool,
) -> List[List[DbObjectAction]]:
    """
    Order a list of object actions with internal dependencies.
    """
    if len(object_action_list_to_be_ordered) == 0:
        return []
    dep_client = copy.deepcopy(
        dependency_client
    )  # this function will make changes to dep_client, but the original dependency_client should not be affected
    filter_mode = "meta_deployment" if is_meta_deployment else "target_deployment"
    dep_client.filter_graph_by_object_ids(
        [a.id for a in object_action_list_to_be_ordered], filter_mode
    )
    dep_client.build_subgraphs()
    ordered_solution_objects = dep_client.get_ordered_objects()
    ordered_solution_objects.sort(key=len)
    ordered_solution_objects.reverse()

    action_type = DbActionType.ADD if is_meta_deployment else DbActionType.ALTER
    object_actions_from_dependency_parser = (
        _generate_actions_from_ordered_solution_objects(
            object_action_list_to_be_ordered, ordered_solution_objects, action_type
        )
    )

    if not _check_all_actions_are_present(
        object_action_list_to_be_ordered, object_actions_from_dependency_parser
    ):
        raise ValueError(
            "Deployment parser sanity check failed: Not all actions appear in ordered action list"
        )

    return object_actions_from_dependency_parser


def _generate_actions_from_ordered_solution_objects(
    unordered_action_list: List[DbObjectAction],
    ordered_solution_object_list: List[List[SolutionObject]],
    action_type: DbActionType,
):
    """
    Given a list of (unordered) actions, order the given list using a nested list of solution objects.
    If an action in the list exists that corresponds to the solution object, use that action.
    If a matching action does not exists, create one.
    """
    log.info("MATCH actions to ordered solution objects")
    ordered_action_list = []

    unordered_action_dict_by_id = defaultdict(list)
    for action in unordered_action_list:
        unordered_action_dict_by_id[action.id].append(action)

    for ordered_solution_sublist in ordered_solution_object_list:
        tmp = []
        for solution_object in ordered_solution_sublist:
            action = _match_action_to_solution_object(
                unordered_action_dict_by_id, solution_object.id
            )
            if action:
                tmp.append(action)
            else:
                tmp.append(
                    DbObjectAction.factory_from_solution_object(
                        solution_object, action_type
                    )
                )
        ordered_action_list.append(tmp)
    return ordered_action_list


def _match_action_to_solution_object(action_dict_by_id, solution_object_id):
    """
    Find a matching action in action_dict_by_id for the given solution_object id.
    If such an action does not exist in action_dict_by_id, return None.

    action_dict_by_id must be a dictionary where each key is an id and each value is a
    list of actions with zero or one elements.
    """
    result = action_dict_by_id[solution_object_id]
    if len(result) == 0:
        log.debug(
            f"OBJECT with id [ '{solution_object_id}' ] NOT in FOUND in action list"
        )
        return None
    elif len(result) > 1:
        raise ValueError(
            f"OBJECT with id [ '{solution_object_id}' ] NOT UNIQUE in action list. Found [ '{len(result)}' ]"
        )
    else:
        return result[0]


def _check_all_actions_are_present(action_list1, action_list2):
    """
    Sanity check after ordering of dependency graph.
    Return True if all actions in action_list1 appear in action_list2.
    Else return False.
    """
    action_list2_ids = [a.id for a in dict_and_list_util.flatten(action_list2)]
    for action1 in dict_and_list_util.flatten(action_list1):
        if not action1.id in action_list2_ids:
            return False
    return True
